/****************************************************************************
 *
 * Portable Wireless Keyboard(TM) add-on driver for linux-Xwindow. v0.3.3
 * Copyright GPL Licence. 
 * 2005.05.14 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

struct keymapping {
    unsigned char hwcode;       /* Input keycode from IR Keyboard to Linux
				   IrDA port( = this program) */
    unsigned int xcode_nml;     /* Output keycode from this program to X
				   server according to xmodmap */ 
    unsigned int xcode_lfn;     /* xcode with Left-Fn key */
    unsigned int xcode_rfn[3];  /* xcode with Right-Fn key */
};

/* This is default keymapping of debian woody */ 
struct keymapping KeyTable[51] = {
    {0x58, 50, 0, {0}},        /* Shift-L Shift_L:none none none */
    {0x09, 24, 10, {50,10}},   /* Q q:Q 1 exclam */
    {0x0a, 25, 11, {50,11}},   /* W w:W 2 at */
    {0x0b, 26, 12, {50,12}},   /* E e:E 3 numbersign */
    {0x11, 38, 0, {0}},        /* A a:A none none */ 
    {0x03, 52, 0, {0}},        /* Z z:Z none none */
    {0x08, 9, 0, {0}},         /* Cmd Escape:none none none */
    {0x1b, 23, 0, {0}},        /* Tab Tab:ISO_Left_Tab none none */
    {0x0c, 27, 13, {50,13}},   /* R r:R 4 dollar */
    {0x12, 39, 0, {0}},        /* S s:S none none */
    {0x10, 53, 0, {0}},        /* X x:X none none */
    {0x18, 66, 0, {0}},        /* Cap_Lk Caps_Lock:none none none */
    {0x1a, 37, 0, {0}},        /* Ctrl Control_L:none none none */
    {0x0d, 28, 14, {50,14}},   /* T t:T 5 percent */
    {0x13, 40, 0, {0}},        /* D d:D none none */
    {0x2c, 54, 0, {0}},        /* C c:C none none */
    {0x2d, 55, 0, {0}},        /* V v:V none none */
    {0x2e, 56, 0, {0}},        /* B b:B none none */
    {0x23, 64, 0, {0}},        /* Alt Alt_L:Meta_L none none */
    {0x14, 41, 0, {0}},        /* F f:F none none */
    {0x17, 65, 0, {0}},        /* Space-L space:none none none */
    {0x16, 43, 0, {0}},        /* H h:H none none */
    {0x15, 42, 0, {0}},        /* G g:G none none */
    {0x0e, 29, 15, {50,15}},   /* Y y:Y 6 asciicircum */
    {0x22, 0, 0, {0}},         /* Fn-L none:none none none */
    {0x37, 65, 0, {0}},        /* Space-R space:none none none */
    {0x51, 100, 0, {97}},      /* Left Left:none none Home */
    {0x2f, 57, 0, {0}},        /* N n:N none none */
    {0x44, 44, 0, {0}},        /* J j:J none none */
    {0x24, 0, 0, {0}},         /* Fn-R none:none none none */
    {0x52, 104, 0, {105}},     /* Down Down:none none Next */
    {0x48, 98, 0, {99}},       /* Up Up:none none Prior */
    {0x4c, 58, 0, {0}},        /* M m:M none none */
    {0x45, 45, 0, {0}},        /* K k:K none none */
    {0x3c, 30, 16, {50,16}},   /* U u:U 7 ampersand */
    {0x53, 102, 0, {103}},     /* Right Right:none none End */
    {0x4d, 59, 0, {0}},        /* Comma comma:less none none */
    {0x47, 47, 0, {0}},        /* Semicolon semicolon:colon none none */
    {0x46, 46, 0, {0}},        /* L l:L none none */
    {0x3d, 31, 17, {50,17}},   /* I i:I 8 asterisk */
    {0x50, 107, 0, {0}},       /* Del Delete:none none none */
    {0x4e, 60, 0, {0}},        /* Period period:greater none none */
    {0x56, 48, 49, {50,49}},   /* Apostrophe apostrophe:quotedbl grave asciitilde */
    {0x38, 20, 34, {50,34}},   /* Minus minus:underscore bracketleft braceleft */
    {0x3e, 32, 18, {50,18}},   /* O o:O 9 parenleft */
    {0x3a, 61, 51, {50,51}},   /* Slash slash:question backslash bar */
    {0x57, 36, 0, {0}},        /* Enter return:none none none */
    {0x32, 22, 0, {0}},        /* BackSp BackSpace:Terminate_Server none none */
    {0x39, 21, 35, {50,35}},   /* Equal equal:plus bracketright braceright */
    {0x3f, 33, 19, {50,19}},   /* P p:P 0 parenright */
    {0x59, 62, 0, {0}}         /* Shift-R Shift-R:none none none */
};

int fd;
unsigned int auto_off_time = 300;    /* seconds */
unsigned int typematic_rate = 1;     /* n times of unit cycle */
unsigned int typematic_delay = 3;    /* n times of unit cycle */
unsigned int num_rptcode = 0;
unsigned char prev_key = 0;
unsigned int lfn_FLAG = 0;
unsigned int rfn_FLAG = 0;
Display *d;

void alarm_inthandler(int fd) {
    close(fd);
    exit(EXIT_SUCCESS);
}
    
void read_rcfile() {
    int i;
    FILE *fp;
    char dmy[255];
    
    fp = fopen("/etc/pwkrc", "r");
    if ( fp == NULL ) {
	fprintf(stderr,"Cannot open /etc/pwkrc.\n");
	fprintf(stderr,"Internal key definition is applied.\n");
	return;
    }

    while ( (fgets(dmy, sizeof(dmy), fp)) != NULL ) {

	if ( dmy[0] == '#' ) continue;
	else if ( dmy[0] == 'd' ) {
	    sscanf(dmy, "%*s %d", &typematic_delay);
	} else if ( dmy[0] == 'r' ) {
	    sscanf(dmy, "%*s %d", &typematic_rate);
	} else if ( dmy[0] == 'a' ) {
	    sscanf(dmy, "%*s %d", &auto_off_time);
	} else if ( isdigit( dmy[0] ) != 0 ) {
	    for ( i=0; i <= 50; i++) {
		sscanf(dmy,"%x %ud %ud %ud:%ud:%ud", 
		       &KeyTable[i].hwcode, 
		       &KeyTable[i].xcode_nml, 
		       &KeyTable[i].xcode_lfn, 
		       &KeyTable[i].xcode_rfn[0], 
		       &KeyTable[i].xcode_rfn[1],
		       &KeyTable[i].xcode_rfn[2]);
		fgets(dmy, sizeof(dmy), fp);
	    } 
	}
    }
    fclose(fp);
}

void translate( unsigned char ch) {

    int  i, j;
    unsigned int *ptr;

    unsigned char msb;   /* 0 : key pressed */
                         /* 1 means key released. */

    msb =  (ch & 128) >> 7 ;   
    ch &= 127 ; 

    for (i = 0; i <= 50 ; i++) {

	if ( ch == KeyTable[i].hwcode ) {
	    if ( ch == 0x22 ) {         /* Left-Fn */
		lfn_FLAG = (1^msb);
		break;
	    } else if ( ch == 0x24 ) {  /* Right-Fn */
		rfn_FLAG = (1^msb);
		break;
	    } else if ( lfn_FLAG == 1 ) {
		if ( ch == 0x50 ) {          /* LeftFn-Del */
		    close(fd);
		    exit(EXIT_SUCCESS);
		}
		XTestFakeKeyEvent(d, KeyTable[i].xcode_lfn, (1^msb), 0);
		XFlush(d);
		break;
	    } else if ( rfn_FLAG == 1 ) {

		ptr = &KeyTable[i].xcode_rfn[0];
		    
		if ( msb == 0 ) {  /* if RFNed make, send proper order */   
		    while ( *ptr != 0 ) {
			XTestFakeKeyEvent(d, *ptr++, (1^msb), 0);
			XFlush(d);
		    }
		    break;
		} else {  /* if RFNed break, send xcodes in reverse order */

		    for ( j=0 ; *ptr != 0; j++) ptr++;
		    for ( --j ; j >= 0; j-- ) {
			XTestFakeKeyEvent(d, KeyTable[i].xcode_rfn[j], (1^msb), 0);
			XFlush(d);
		    }
		    break;
		}		    
	    } else {
		XTestFakeKeyEvent(d, KeyTable[i].xcode_nml, (1^msb), 0);
		XFlush(d);
		break;
	    }
	}
    }
}

int process_hwcode( unsigned char ch) {
	
    if ( ch == 0x5c ) {                           /* key repetition code */
	if (  ((prev_key & 128) >> 7) == 0  ) {   /* if prev is make */
	    num_rptcode++;
	    if ( num_rptcode >= typematic_delay ) {
		translate(prev_key);
		num_rptcode--;
	    } 
	} /* if prev is break key, ignore. (any break key cannot repeat) */
	return 1;          
    }
       
    if ( ch == prev_key ) {
	return 1;          /* ignore one of the two duplicated break code */
    }

    translate( ch );
    prev_key = ch;
    return 1;
}
	
int main(int argc, char *argv[]) {

    unsigned char ch, hwcode;
    int nread;
    char *device = "/dev/ttyS1";

    struct termios oldtio, newtio;
    struct sigaction newhandler;
    sigset_t blocked;

    /* signal install */
    newhandler.sa_handler = alarm_inthandler;
    sigemptyset(&blocked);
    sigaddset(&blocked, SIGALRM);
    newhandler.sa_mask = blocked;
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
	perror("sigaction install error\n");
	exit(EXIT_FAILURE);
    }

    /* Opening irda port device */
    if ( argc >= 2 ) device = argv[1]; 
    fd = open(device, O_RDONLY | O_NOCTTY );
    if (fd == -1)  {
    	fprintf(stderr, "Cannot open %s\n", device);
    	exit(EXIT_FAILURE);
    }

    read_rcfile();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(fd, &oldtio); 
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
               
    d = XOpenDisplay(NULL);
    
    while (1) {
	alarm( auto_off_time );

	nread = read(fd, &ch, 1);
        if ( ( nread > 0 ) && ( ch == 0xff ) ) {
      
	    nread = read(fd, &ch, 1);
	    if ( (nread > 0) && ( ch == 0xc0 ) ) {
	    
		nread = read(fd, &hwcode, 1);
		read(fd, &ch, 1);
		read(fd, &ch, 1); 
		nread = read(fd, &ch, 1);

		if ( ch == 0xc1 ) process_hwcode(hwcode);		
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n");
	    fprintf(stderr,"Recommend removing other irda modules.\n");
	    exit(EXIT_FAILURE);
	}         
    }  
}
